import socket
import struct

from gbvision.constants.net import TCP_HEADERS_STRUCT
from gbvision.exceptions.tcp_stream_closed import TCPStreamClosed
from .stream_receiver import StreamReceiver


class TCPStreamReceiver(StreamReceiver):
    """
    This class uses TCP to receive a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the server and the receiver is the client

    :param ip: The IPv4 address of the stream broadcaster, for example '10.45.90.8'
    :param port: The TCP port to use
    """

    def __init__(self, ip: str, port: int, *args, **kwargs):
        StreamReceiver.__init__(self, *args, **kwargs)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_addr = (ip, port)
        self.socket.connect(self.server_addr)
        self.headers_size = struct.calcsize(TCP_HEADERS_STRUCT)

    def _receive(self):
        try:
            return self.socket.recv(2 ** 32)
        except OSError as e:
            self.release()
            raise TCPStreamClosed() from e

    def _get_bytes(self) -> bytes:
        data = b''
        while len(data) < self.headers_size:
            data += self._receive()
        packed_msg_size = data[:self.headers_size]
        data = data[self.headers_size:]
        msg_size = struct.unpack(TCP_HEADERS_STRUCT, packed_msg_size)[0]
        while len(data) < msg_size:
            data += self._receive()
        frame_data = data[:msg_size]
        return frame_data

    def release(self) -> None:
        self.socket.close()

    def is_opened(self) -> bool:
        return self.socket.fileno() != -1



from .object_finder import ShapeFinder
from gbvision.utils.shapes.base_polygon import BasePolygon
from ..shapes.base_shape import BaseShapeType


class PolygonFinder(ShapeFinder):
    """
    Finds any generic polygon, not recommended when another finder can be used
    """

    @staticmethod
    def _base_shape() -> BaseShapeType:
        return BasePolygon



import cv2

from .base_contour import BaseContour
from gbvision.constants.types import Polygon, Contour


class BasePolygon(BaseContour):
    @staticmethod
    def from_contour(cnt: Contour, arc_length_multiplier=0.05) -> Polygon:
        arc_length = arc_length_multiplier * cv2.arcLength(cnt, True)
        return cv2.approxPolyDP(cnt, arc_length, True)



from .object_finder import ShapeFinder
from ..shapes.base_contour import BaseContour
from ..shapes.base_shape import BaseShapeType


class ContourFinder(ShapeFinder):
    """
    Finds any generic shape, not recommended when another finder can be used
    """

    @staticmethod
    def _base_shape() -> BaseShapeType:
        return BaseContour



import cv2

from gbvision.constants.types import Frame, Color, Coordinates, Number


def draw_text(frame: Frame, text: str, coords: Coordinates, font_scale: Number, color: Color,
              font=cv2.FONT_HERSHEY_SIMPLEX, *args, **kwargs) -> Frame:
    """
    Draws the text on a copy of the frame and returns the copy

    :param frame: The frame to draw on
    :param text: The text to draw
    :param coords: The coordinates of the bottom-left corner of the text
    :param font_scale: The size of the drawn text (multiplied by the default size of the font)
    :param color: The color to draw the text in
    :param font: The font, an opencv font constant, default is cv2.FONT_HERSHEY_SIMPLEX
    :param args: Additional arguments to cv2.putText
    :param kwargs: Additional keyword arguments to cv2.putText (such as thickness)
    :return: A copy of the frame with the text drawn on it
    """
    frame = frame.copy()
    cv2.putText(frame, text, coords, font, font_scale, color, *args, **kwargs)
    return frame



from .object_finder import ShapeFinder
from gbvision.utils.shapes.base_circle import BaseCircle
from ..shapes.base_shape import BaseShapeType


class CircleFinder(ShapeFinder):
    """
    Finds specific circular shaped object, and performs distance transformation
    """

    @staticmethod
    def _base_shape() -> BaseShapeType:
        return BaseCircle



import socket
import struct

from gbvision.constants.net import LOCAL_SERVER_IP, TCP_HEADERS_STRUCT
from gbvision.exceptions.tcp_stream_closed import TCPStreamClosed
from .stream_broadcaster import StreamBroadcaster


class TCPStreamBroadcaster(StreamBroadcaster):
    """
    This class uses TCP to send a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the server and the receiver is the client
    
    :param port: The TCP port to use
    """

    def __init__(self, port: int, *args, **kwargs):
        StreamBroadcaster.__init__(self, *args, **kwargs)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_addr = (LOCAL_SERVER_IP, port)
        self.socket.bind(self.server_addr)
        self.socket.listen(10)
        self.socket, addr = self.socket.accept()

    def _send_bytes(self, data):
        try:
            data = struct.pack(TCP_HEADERS_STRUCT, len(data)) + data
            self.socket.send(data)
        except IOError as e:
            raise TCPStreamClosed() from e
        self._update_time()

    def release(self) -> None:
        self.socket.close()

    def is_opened(self) -> bool:
        return self.socket.fileno() != -1



from .async_stream_receiver import AsyncStreamReceiver
from .fragmented_udp_stream_receiver import FragmentedUDPStreamReceiver


class AsyncFragmentedUDPStreamReceiver(AsyncStreamReceiver, FragmentedUDPStreamReceiver):
    def __init__(self, port, *args, **kwargs):
        FragmentedUDPStreamReceiver.__init__(self, port, *args, **kwargs)
        AsyncStreamReceiver.__init__(self, *args, **kwargs)

    def _read(self):
        return FragmentedUDPStreamReceiver.read(self)

    def _release(self) -> None:
        FragmentedUDPStreamReceiver.release(self)



from .object_finder import ShapeFinder
from gbvision.utils.shapes.base_rotated_rect import BaseRotatedRect
from ..shapes.base_shape import BaseShapeType


class RotatedRectFinder(ShapeFinder):
    """
    Finds a rectangular object, but rotated.
    Recommended to use when you know the shape isn't parallel to the camera
    """

    @staticmethod
    def _base_shape() -> BaseShapeType:
        return BaseRotatedRect



from copy import deepcopy
from typing import List

import numpy as np
import cv2

from .base_shape import BaseShape
from .base_rect import BaseRect
from gbvision.constants.types import Contour, Number, Point, Rect, Frame, Color
from gbvision.constants.math import EPSILON


class BaseContour(BaseShape):
    @classmethod
    def set_center(cls, shape: Contour, new_center: Point) -> Contour:
        current_center = cls.center(shape)
        return np.array(shape) + [new_center[0] - current_center[0], new_center[1] - current_center[1]]

    @classmethod
    def _unsafe_draw(cls, frame: Frame, shape: Contour, color: Color, *args, **kwargs) -> None:
        return cls._unsafe_draw_multiple(frame, [shape], color, *args, **kwargs)

    @classmethod
    def _unsafe_draw_multiple(cls, frame: Frame, shapes: List[Contour], color: Color, *args, **kwargs) -> None:
        return cv2.drawContours(frame, shapes, -1, color, *args, **kwargs)

    @staticmethod
    def to_bounding_rect(shape: Contour) -> Rect:
        return BaseRect.from_contour(shape)

    @staticmethod
    def from_bounding_rect(bounding_rect: Rect) -> Contour:
        return np.array([[bounding_rect[0], bounding_rect[1]],
                         [bounding_rect[0] + bounding_rect[2], bounding_rect[1]],
                         [bounding_rect[0] + bounding_rect[2], bounding_rect[1] + bounding_rect[3]],
                         [bounding_rect[0], bounding_rect[1] + bounding_rect[3]]])

    @staticmethod
    def from_contour(cnt: Contour) -> Contour:
        return deepcopy(cnt)

    @staticmethod
    def area(shape: Contour) -> Number:
        return cv2.contourArea(shape)

    @staticmethod
    def center(shape: Contour) -> Point:
        m = cv2.moments(shape)
        return int(m['m10'] / (m['m00'] + EPSILON)), int(m['m01'] / (m['m00'] + EPSILON))

    @staticmethod
    def collision(shape1: Contour, shape2: Contour) -> bool:
        return NotImplemented




import cv2

from .base_rotated_rect import BaseRotatedRect
from gbvision.constants.types import Ellipse, Frame, Color


class BaseEllipse(BaseRotatedRect):

    @staticmethod
    def _unsafe_draw(frame: Frame, shape: Ellipse, color: Color, *args, **kwargs) -> None:
        cv2.ellipse(frame, shape, color, *args, **kwargs)

    @staticmethod
    def collision(shape1: Ellipse, shape2: Ellipse) -> bool:
        return NotImplemented

    from_contour = cv2.fitEllipse



import abc
import time
from threading import Thread
from typing import Tuple

from gbvision.constants.types import Frame, Number
from gbvision.utils.readable import Readable


class AsyncReadable(Readable, abc.ABC):
    """
    An async readable class that constantly reads on another thread
    """

    def __init__(self):
        self.__ok, self.__frame = False, None
        self.__is_running = True
        self.__thread = Thread(target=self.__async_read_wrapper)
        self.__thread.start()

    def read(self) -> Tuple[bool, Frame]:
        return self.__ok, self.__frame

    def has_started_reading(self) -> bool:
        """
        Checks if the async thread has started reading

        :return: True if the async thread started, False otherwise
        """
        return self.__ok

    def wait_start_reading(self, wait_time: Number = 0.01) -> None:
        """
        Waits until the async thread starts reading

        :param wait_time: Amounts of seconds to wait in every check interval, default is 0.01
        """
        while not self.has_started_reading():
            time.sleep(wait_time)

    @abc.abstractmethod
    def _read(self) -> Tuple[bool, Frame]:
        """
        Reads from the camera synchronously (similar to Readable.read), unsafe, not to use by the programmer
        this method will usually simply call super(self, ReadableClass).read()

        :return: Tuple of bool (indicates if read was successful) and the frame (if successful, else None)
        """

    def __async_read_wrapper(self) -> None:
        while self.__is_running:
            self.__ok, self.__frame = self._read()

    def release(self) -> None:
        self.__is_running = False
        self.__thread.join()
        self._release()

    @abc.abstractmethod
    def _release(self) -> None:
        """
        Unsafely releases this readable's resources without killing the reading thread
        """


from typing import Tuple, List, Union, Callable, TypeVar, Optional, Any
from numpy import ndarray

Number = Union[int, float]
Point = Tuple[Number, Number]  # (x, y)
Contour = ndarray  # [[x0, y0], [x1, y1], ... ]
Circle = Tuple[Tuple[Number, Number], Number]  # ((center_x, center_y), radius)
Rect = Tuple[Number, Number, Number, Number]  # (x, y, width, height)
Polygon = Union[List[Point], Contour]
FixedPolygon = List[Point]
RotatedRect = Tuple[Point, Point, float]  # ((center_x, center_y), (width, height), angle in degrees)
Ellipse = RotatedRect
Frame = Optional[ndarray]
Color = Tuple[int, int, int]
FilterFunction = Callable[[Frame], Frame]
Coordinates = Tuple[int, int]  # (x, y)
Line = Tuple[Point, Point]
Location = Union[Tuple[Number, Number, Number], ndarray]
ROI = Tuple[int, int, int, int]  # (x, y, width, height)
Shape = TypeVar('Shape')
ColorThresholdParams = Tuple[Tuple[Number, Number], Tuple[Number, Number], Tuple[Number, Number]]
# ((channel0 low, channel0 high), (channel1 low, channel1 high), (channel2 low, channel2 high))
GrayScaleThresholdParams = Tuple[Number, Number]  # (low, high)

try:
    from typing import Protocol


    class TrackerType(Protocol):
        init: Callable[[Frame, Rect], bool]
        update: Callable[[Frame], Tuple[bool, Rect]]
except ImportError:
    # Protocol is not supported in python 3.7
    TrackerType = Any



import numpy as np
import cv2

from .base_shape import BaseShape
from .base_convex_polygon import BaseConvexPolygon
from gbvision.constants.types import RotatedRect, Number, Point, Rect, Frame, Color


class BaseRotatedRect(BaseShape):
    @staticmethod
    def set_center(shape: RotatedRect, new_center: Point) -> RotatedRect:
        return new_center, shape[1], shape[2]

    @staticmethod
    def _unsafe_draw(frame: Frame, shape: Rect, color: Color, *args, **kwargs) -> None:
        box = cv2.boxPoints(shape)
        box = np.int0(box)
        cv2.drawContours(frame, [box], 0, color, *args, **kwargs)

    @staticmethod
    def to_bounding_rect(shape: RotatedRect) -> Rect:
        rotated_rect = shape
        x = rotated_rect[0][0]
        y = rotated_rect[0][1]
        w = rotated_rect[1][0]
        h = rotated_rect[1][1]
        a = np.deg2rad(rotated_rect[2])

        bound_w = w * np.cos(a) + h * np.sin(a)
        bound_h = h * np.cos(a) + w * np.sin(a)
        bound_x = x - bound_w / 2
        bound_y = y - bound_h / 2
        return bound_x, bound_y, bound_w, bound_h

    @staticmethod
    def from_bounding_rect(bounding_rect: Rect) -> RotatedRect:
        return ((bounding_rect[0] - bounding_rect[2]) / 2, (bounding_rect[1] + bounding_rect[3] / 2)), \
               bounding_rect[2:4], 0

    from_contour = cv2.minAreaRect

    @staticmethod
    def center(shape: RotatedRect) -> Point:
        return shape[0]

    @staticmethod
    def collision(shape1: RotatedRect, shape2: RotatedRect) -> bool:
        return BaseConvexPolygon.collision(cv2.boxPoints(shape1), cv2.boxPoints(shape2))

    @classmethod
    def area(cls, shape: RotatedRect) -> Number:
        return shape[1][0] * shape[1][1]



import abc
from typing import Type, List

import numpy as np

from gbvision.constants.types import Shape, Number, Rect, Point, Contour, Frame, Color


class BaseShape(abc.ABC):
    def __init__(self):
        raise Exception("Can't create an instance of a base shape class")

    @staticmethod
    @abc.abstractmethod
    def to_bounding_rect(shape: Shape) -> Rect:
        """
        Finds the bounding rect of this shape

        :param shape: The shape
        :return: The bounding rect of this shape, as a gbvision.Rect
        """

    @staticmethod
    @abc.abstractmethod
    def from_bounding_rect(bounding_rect: Rect) -> Shape:
        """
        Converts a rect to the closest possible object of this shape that is contained in it

        :param bounding_rect: The rect to convert
        :return: A shape of this type that is as close as possible to this rect
        """

    @staticmethod
    @abc.abstractmethod
    def from_contour(cnt: Contour) -> Shape:
        """
        Converts a single contour to this shape

        :param cnt: The contour
        :return: A shape representing this contour
        """

    @classmethod
    def from_contours(cls, cnts: List[Contour]) -> List[Shape]:
        """
        Converts the given list of contours to a list of this shape

        :param cnts: The contours to convert
        :return: A list of this shape, the same length as the contours list
        """
        return list(map(cls.from_contour, cnts))

    @staticmethod
    @abc.abstractmethod
    def collision(shape1: Shape, shape2: Shape) -> bool:
        """
        checks if the two shapes are colliding

        :param shape1: the first shape
        :param shape2: the second shape
        :return: True if the shapes are colliding, False otherwise
        """

    @staticmethod
    @abc.abstractmethod
    def area(shape: Shape) -> Number:
        """
        calculates the area of the shape

        :param shape: the shape
        :return: the area of the shape
        """

    @staticmethod
    @abc.abstractmethod
    def center(shape: Shape) -> Point:
        """
        calculates the center-of-mass of the shape

        :param shape: the shape
        :return: the center of the shape
        """

    @classmethod
    def root_area(cls, shape: Shape) -> Number:
        """
        calculates the square root of the area of the shape
        default is the square root of cls.shape_area, but it can be overridden in case there is a simpler way\
         (for example for circles)

        :param shape: the shape
        :return: the square root of the area of the shape
        """
        return np.sqrt(cls.area(shape))

    @classmethod
    def sort(cls, shapes: List[Shape]) -> List[Shape]:
        """
        sorts the list of shapes by area, should be overridden to use area root in case it's better
        doesn't modify the given list, returns a new list

        :param shapes: the list of shapes to sort
        :return: a sorted copy of the list of shapes
        """
        return sorted(shapes, key=cls.area, reverse=True)

    @classmethod
    def filter_inners(cls, shapes: List[Shape]) -> List[Shape]:
        """
        filters out all shapes that are colliding with a shape with a smaller index in the given list
        returns a list of all shapes that aren't colliding

        :param shapes: the list of shapes
        :return: the filtered list of shapes
        """
        filtered_shapes = []
        for i, shape in enumerate(shapes):
            shape_invalid = False
            for j in range(i):
                shape_invalid = cls.collision(shape, shapes[j])
                if shape_invalid:
                    break
            if not shape_invalid:
                filtered_shapes.append(shape)
        return filtered_shapes

    @classmethod
    def from_contours_sorted(cls, cnts: List[Contour]) -> List[Shape]:
        """
        Converts the given list of contours to a list of this shape, and sorts them by size

        :param cnts: The contours to convert
        :return: A list of this shape, the same length as the contours list
        """
        return cls.sort(cls.from_contours(cnts))

    @classmethod
    def distance_squared(cls, shape1: Shape, shape2: Shape) -> Number:
        """
        Finds the square of the distance between the centers of the shapes

        :param shape1: The first shape
        :param shape2: The second shape
        :return: The square of distance between the two centers
        """
        center1 = cls.center(shape1)
        center2 = cls.center(shape2)
        return (center1[0] - center2[0]) ** 2 + (center1[1] - center2[1]) ** 2

    @classmethod
    def distance(cls, shape1: Shape, shape2: Shape) -> Number:
        """
        Finds the square of the distance between the centers of the shapes

        :param shape1: The first shape
        :param shape2: The second shape
        :return: The square of distance between the two centers
        """
        return np.sqrt(cls.distance_squared(shape1, shape2))

    @staticmethod
    @abc.abstractmethod
    def _unsafe_draw(frame: Frame, shape: Shape, color: Color, *args, **kwargs) -> None:
        """
        Unsafely draws this shape on the frame
        This method should modify the given frame and not return anything

        :param frame: The frame to draw on
        :param shape: The shape to draw
        :param color: The color to draw with
        :param args: Optional arguments to pass to the drawing function
        :param kwargs: Optional keyword arguments to pass to the drawing function
        """

    @classmethod
    def _unsafe_draw_multiple(cls, frame: Frame, shapes: List[Shape], color: Color, *args, **kwargs) -> None:
        """
        Unsafely draws this object on the frame
        This method should modify the given frame and not return anything

        :param frame: The frame to draw on
        :param shapes: The list of shapes to draw
        :param color: The color to draw with
        :param args: Optional arguments to pass to the drawing function
        :param kwargs: Optional keyword arguments to pass to the drawing function
        """
        for shape in shapes:
            cls._unsafe_draw(frame, shape, color, *args, **kwargs)

    @classmethod
    def draw(cls, frame: Frame, shape: Shape, color: Color, *args, **kwargs) -> Frame:
        """
        Draws the given shape on the frame
        This method does not change to original frame, and instead creates a copy of it and returns the copy

        :param frame: The frame to draw on
        :param shape: The shape to draw
        :param color: The color to use
        :param args: Optional arguments to pass to the drawing function
        :param kwargs: Optional keyword arguments to pass to the drawing function
        :return: A copy of the frame with the shape drawn on it
        """
        frame = frame.copy()
        cls._unsafe_draw(frame, shape, color, *args, **kwargs)
        return frame

    @staticmethod
    @abc.abstractmethod
    def set_center(shape: Shape, new_center: Point) -> Shape:
        """
        Returns an identical copy that has been moved to a new location
        Only the center should be different

        :param shape: The shape
        :param new_center: The new center, calling BaseShape.center on the return value should result in this
        :return: A copy of this shape, with the new center
        """

    @classmethod
    def draw_multiple(cls, frame: Frame, shapes: List[Shape], color: Color, *args, **kwargs) -> Frame:
        """
        Draws the given shapes on the frame
        This method does not change to original frame, and instead creates a copy of it and returns the copy

        :param frame: The frame to draw on
        :param shapes: The list of shapes to draw
        :param color: The color to use
        :param args: Optional arguments to pass to the drawing function
        :param kwargs: Optional keyword arguments to pass to the drawing function
        :return: A copy of the frame with the shape drawn on it
        """
        frame = frame.copy()
        cls._unsafe_draw_multiple(frame, shapes, color, *args, **kwargs)
        return frame

    @classmethod
    def sort_and_filter_inners(cls, shapes: List[Shape]):
        """
        Sorts the given list of shapes and filters out any inner shapes

        :param shapes: The list of shapes
        :return: The list of shapes sorted, with all inner shapes filtered out
        """
        return cls.filter_inners(cls.sort(shapes))


BaseShapeType = Type[BaseShape]



from abc import ABC

import cv2

from gbvision.models.system import EMPTY_PIPELINE
from .opencv_window import OpenCVWindow
from .wrapper_window import WrapperWindow
from gbvision.constants.types import Frame, ROI


class WrapperOpenCVWindow(OpenCVWindow, WrapperWindow, ABC):
    """
    A basic window that displays a feed from a camera

    :param wrap_object: The object from which to read the frame
    :param drawing_pipeline: Optional. A pipeline of drawing functions that will run on the frame before displaying it
    :param exit_button: An array of keys (a string), when one of the keys are pressed the window will be closed
    :param window_name: The title of the window
    """

    def __init__(self, window_name: str, wrap_object, exit_button='qQ', drawing_pipeline=EMPTY_PIPELINE,
                 flags=cv2.WINDOW_FREERATIO):
        OpenCVWindow.__init__(self, window_name, flags=flags, exit_button=exit_button)
        WrapperWindow.__init__(self, window_name=window_name, wrap_object=wrap_object,
                               drawing_pipeline=drawing_pipeline)

    def show_async(self) -> None:
        self.flags = cv2.WINDOW_AUTOSIZE
        WrapperWindow.show_async(self)

    def select_roi(self, frame: Frame = None) -> ROI:
        if frame is None:
            _, frame = self._get_frame()
        return OpenCVWindow.select_roi(self, frame)



import abc
from typing import Tuple
from .releasable import Releasable

from gbvision.constants.types import Frame


class Readable(Releasable, abc.ABC):
    """
    An interface representing an object you can read frames from, such as cameras and streams
    """

    @abc.abstractmethod
    def read(self) -> Tuple[bool, Frame]:
        """
        Reads from the readable and returns the result

        :return: False, None if the operation was unsuccessful, else True, frame_result
        """

    @abc.abstractmethod
    def get_width(self) -> int:
        """
        :return: The width of the frames read by this readable
        """

    @abc.abstractmethod
    def get_height(self) -> int:
        """
        :return: The height of the frames read by this readable
        """



from copy import deepcopy

import cv2

from .base_shape import BaseShape
from gbvision.constants.types import Rect, Number, Point, Frame, Color


class BaseRect(BaseShape):
    @classmethod
    def set_center(cls, shape: Rect, new_center: Point) -> Rect:
        current_center = cls.center(shape)
        center_diff = new_center[0] - current_center[0], new_center[1] - current_center[1]
        return shape[0] + center_diff[0], shape[1] + center_diff[1], shape[2], shape[3]

    @staticmethod
    def _unsafe_draw(frame: Frame, shape: Rect, color: Color, *args, **kwargs) -> None:
        cv2.rectangle(frame, (int(shape[0]), int(shape[1])), (int(shape[0] + shape[2]), int(shape[1] + shape[3])),
                      color, *args, **kwargs)

    @staticmethod
    def to_bounding_rect(shape: Rect) -> Rect:
        return deepcopy(shape)

    @staticmethod
    def from_bounding_rect(bounding_rect: Rect) -> Rect:
        return deepcopy(bounding_rect)

    from_contour = cv2.boundingRect

    @staticmethod
    def center(shape: Rect) -> Point:
        return shape[0] + (shape[2] / 2), shape[1] + (shape[3] / 2)

    @staticmethod
    def collision(shape1: Rect, shape2: Rect) -> bool:
        return not (shape1[0] > shape2[0] + shape2[2] or
                    shape1[0] + shape1[2] < shape1[0] or
                    shape1[1] > shape2[1] + shape2[3] or
                    shape1[1] + shape1[3] < shape2[1])

    @classmethod
    def area(cls, shape: Rect) -> Number:
        return shape[2] * shape[3]



from typing import Any, Callable, Iterable, Optional
import functools


class PipeLine:
    """
    A class representing a pipeline of function
    each function receives one input, which is the output of the previous function in the pipeline
    pipelines are great for representing a long computer vision function (which is why such functions
    are called pipelines).
    the PipeLine class can also be used as a function decorator
    for example, creating a pipeline that adds 1 to it's output can be done in two ways:
    Example::
        inc = PipeLine(lambda x: x + 1)
        three = inc(2)

    or
    Example::
        @PipeLine
        def inc(x):
            return x + 1
        three = inc(2)


    You can create a PipeLine from multiple functions:
    Example::
        open_and_read_file = PipeLine(open, lambda x: x.read())
        text = open_and_read_file("file.txt")

    You can also inherit from the PipeLine class to make a PipeLine factory:
    Example::
        class Adder(PipeLine):
            def __init__(self, num):
                self.num = num
                PipeLine.__init__(self, self.adding_func)
            def adding_func(self, item):
                return item + self.num

    You can also do it like this:
    Example::
        class Adder(PipeLine):
            def __init__(self, num):
                def adding_func(item):
                    return item + num

                PipeLine.__init__(self, adding_func)

    You can use combine a few PipeLines together to create function composition:
    Example::
        multiply_by_2_then_add_3 = PipeLine(lambda x: x * 2) + PipeLine(lambda x: x + 3)

    :param functions: A tuple of functions to run one after the other as a pipeline
    :param qualname: Optional. A string that will be the __qualname__ of the pipeline object
    :param module: Optional. A string that will be the __module__ of the pipeline object
    """
    __NO_DOCS = '\n\tNo Docs :('

    @staticmethod
    def __get_qual_name(func: Callable[[Any], Any]) -> str:
        return func.__qualname__ if hasattr(func, '__qualname__') else 'Unknown Function'

    def __init__(self, *functions: Callable[[Any], Any], qualname: Optional[str] = None, module: Optional[str] = None):
        self.functions = []
        for f in functions:
            if isinstance(f, PipeLine):
                self.functions += f.functions
            else:
                self.functions.append(f)

        # set qual name
        if qualname is not None:
            self.__qualname__ = qualname
        if len(functions) == 1:
            if hasattr(functions[0], '__qualname__'):
                self.__qualname__ = functions[0].__qualname__
        elif len(functions) > 1:
            self.__qualname__ = ' + '.join(self.__get_qual_name(func) for func in functions) or None

        # set module
        if module is not None:
            self.__module__ = module
        if len(functions) == 1:
            if hasattr(functions[0], '__module__'):
                self.__module__ = functions[0].__module__
        elif len(functions) > 1:
            all_modules = set(x.__module__ for x in functions if hasattr(x, '__module__'))
            if len(all_modules) == 1:
                self.__module__ = next(iter(all_modules))

        # set doc
        if len(functions) == 1:
            self.__doc__ = functions[0].__doc__
        elif len(functions) > 1:
            all_docs = []
            for i, func in enumerate(functions):
                all_docs.append(f'{i}. {self.__get_qual_name(func)}\n{func.__doc__ or self.__NO_DOCS}')
            self.__doc__ = '\n\n'.join(all_docs)

    def __call__(self, image: Any) -> Any:
        """
        Activate this pipeline and return the result

        :param image: The input to the first function in the pipeline (also the input to the entire pipeline) \
        doesn't have to be an image, can be anything
        :return: The output of the last function in the pipeline, can be data type
        """
        return functools.reduce(lambda x, f: f(x), self.functions, image)

    def __add__(self, other: Callable[[Any], Any]) -> 'PipeLine':
        """
        Creates a new pipeline which uses the output of this pipeline as input to the other pipeline

        :param other: The second pipeline
        :return: A new pipeline, and calling this pipeline with the parameter image is similar to \
        performing other(self(image))
        """
        if isinstance(other, PipeLine):
            return PipeLine(*self.functions + other.functions)
        return PipeLine(*self.functions + [other])

    def __radd__(self, other: Callable[[Any], Any]) -> 'PipeLine':
        """
        Adds this PipeLine to another function that isn't a PipeLine

        :param other: The function
        :return: A new PipeLine which performs self(other(image)) on the parameter image
        """
        return PipeLine(other) + self

    def __getitem__(self, item: int) -> Callable[[Any], Any]:
        """
        Gets the function at index item

        :param item: The index
        :return: The item
        """
        return self.functions[item]

    def __setitem__(self, key: int, value: Callable[[Any], Any]) -> None:
        """
        Sets the function at index key to the new function value

        :param key: The index
        :param value: The new function
        """
        self.functions[key] = value

    def __iter__(self) -> Iterable[Callable[[Any], Any]]:
        """
        :return: An iterator that iterates through all functions in this pipeline
        """
        return iter(self.functions)

    def __len__(self):
        """
        :return: The amount of functions in this pipeline
        """
        return len(self.functions)



from typing import Tuple

from .wrapper_opencv_window import WrapperOpenCVWindow
from gbvision.utils.readable import Readable
from gbvision.constants.types import Frame


class ReadableWindow(WrapperOpenCVWindow):
    """
    A basic window that displays the stream from any readable

    :param wrap_object: A readable to read the frames from
    """

    def __init__(self, window_name: str, wrap_object: Readable, *args, **kwargs):
        WrapperOpenCVWindow.__init__(self, window_name, wrap_object, *args, **kwargs)

    def _get_frame(self) -> Tuple[bool, Frame]:
        return self.wrap_object.read()



import abc
import pickle
from typing import Optional

import cv2
import time

from gbvision.utils.releasable import Releasable
from gbvision.constants.math import EPSILON
from gbvision.constants.types import Frame, Coordinates


class StreamBroadcaster(Releasable, abc.ABC):
    """
    An abstract broadcaster that sends stream to a stream receiver
    
    :param shape: Optional, the shape (x, y) of the sent frame, when set to something other then (0, 0) it overrides
        the fx and fy parameters, when set to (0, 0) it is not used
    :param fx: Ratio between width of the given frame to the width of the frame sent, default is 1 (same width)
    :param fy: ratio between height of the given frame to the height of the frame sent, default is 1 (same height)
    :param use_grayscale: boolean indicating if the frame should be converted to grayscale when sent,
        default is False
    :param max_fps: integer representing the maximum fps (frames per second) of the stream, when set to None
        there is no fps limitation, default is None
    :param max_bitrate: Integer that determines the max bitrate of video stream.
        The bitrate is messured with Kbps and default is None.
    """

    def __init__(self, shape: Optional[Coordinates] = None, fx: Optional[float] = None, fy: Optional[float] = None,
                 use_grayscale: bool = False, max_fps: int = None, im_encode='.jpg', max_bitrate: int = None):
        self.shape = shape
        self.fx = fx
        self.fy = fy
        self.use_grayscale = use_grayscale
        self.max_fps = max_fps
        self.prev_time = 0.0
        self.im_encode = im_encode
        self.max_bitrate = max_bitrate

    def send_frame(self, frame: Frame) -> None:
        """
        safely sends the frame, doing all the pre-processing required
        :param frame:
        :return:
        """
        if frame is not None:
            frame = self._prep_frame(frame)
            frame = cv2.imencode(self.im_encode, frame)[1]
        data = self._to_bytes(frame)
        if self._can_send_bytes(data):
            self._send_bytes(data)
            self._update_time()

    @abc.abstractmethod
    def _send_bytes(self, data: bytes) -> None:
        """
        Unsafely sends the given binary formatted frame to the stream receiver
        should not be used by the programmer, only by the API

        :param data: The frame to send in a binary format (like pickle)
        """

    def _prep_frame(self, frame: Frame) -> Frame:
        """
        Prepares an image to be sent
        Resize and convert the colors of the image by the parameters of the stream broadcaster

        :param frame: The frame to prepare
        :return: The frame after preparation
        """
        if self.shape is not None:
            frame = cv2.resize(frame, self.shape)
        if self.fx is not None or self.fy is not None:
            frame = cv2.resize(frame, (0, 0), fx=self.fx or 1.0, fy=self.fy or 1.0)
        if self.use_grayscale and len(frame.shape) > 2:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        return frame

    @staticmethod
    def _to_bytes(frame: object) -> bytes:
        """
        Converts the given python object to a binary representation

        :param frame: The python object to parse
        :return: A binary representation of the python object
        """
        return pickle.dumps(frame)

    def _legal_time(self) -> bool:
        """
        Checks if at the fps will not pass the max fps limit if an image will be sent at the current moment
        
        :return: True if the image can be sent, False otherwise
        """
        return self.max_fps is None or (time.time() - self.prev_time) * self.max_fps >= 1

    def _update_time(self) -> None:
        """
        Updates the previous time a frame was sent, used at the end of send_frame
        """
        self.prev_time = time.time()

    def _legal_bitrate(self, data: bytes) -> None:
        """
        Checks if sending the data will be a violation of the bitrate limit

        :param data: The data to send
        :return: True if there's no bitrate limit or frame bitrate is below max bitrate, False otherwise
        """
        return self.max_bitrate is None or len(data) / (
                (time.time() - self.prev_time + EPSILON) * 1000) <= self.max_bitrate

    def _can_send_bytes(self, data: bytes) -> bool:
        """
        Checks if the data can be sent under the provided limitations (max FPS, max KBPS and etc)

        :param data: The data to send
        :return: True if sending the data will not violate the limitations, False otherwise
        """
        if not self._legal_time():
            return False

        if not self._legal_bitrate(data):
            return False

        return True



import abc
from typing import Union

import numpy as np

from gbvision.constants.types import Number
from gbvision.utils.readable import Readable
from .camera_data import CameraData


class Camera(Readable, abc.ABC):
    """
    An abstract class representing a camera
    """

    @abc.abstractmethod
    def is_opened(self) -> bool:
        """
        Checks if the camera can be read from

        :return: True if the camera can be read from, otherwise False
        """

    @abc.abstractmethod
    def set_exposure(self, exposure: Union[int, float, bool]) -> bool:
        """
        Sets the camera's exposure

        :param exposure: the new exposure
        :return: True on success, False on failure
        """

    @abc.abstractmethod
    def set_auto_exposure(self, auto: Union[int, float, bool]) -> bool:
        """
        Sets the camera's auto exposure

        :param auto: the new auto exposure
        :return: True on success, False on failure
        """

    @abc.abstractmethod
    def get_data(self) -> CameraData:
        """
        :return: this camera's constant descriptor (must be the real descriptor, can't be a copy) \
            when the values of this descriptor are changed, the values of the real camera descriptor must also change
        """

    @abc.abstractmethod
    def get_width(self) -> int:
        """
        :return: The width of a frame read from this camera
        """

    @abc.abstractmethod
    def get_height(self) -> int:
        """
        :return: The height of a frame read from this camera
        """

    @abc.abstractmethod
    def _set_width(self, width: int) -> None:
        """
        Snsafe set width
        Only used by rescale, resize and set_frame size methods
        Never to be used by the programmer, only by the API

        :param width: new width
        """

    @abc.abstractmethod
    def _set_height(self, height: int) -> None:
        """
        Unsafe set height
        Only used by rescale, resize and set_frame size methods
        Never to be used by the programmer, only by the API

        :param height: new height
        """

    @abc.abstractmethod
    def get_fps(self) -> Number:
        """
        Gets the fps of this camera

        :return: The fps of the camera
        """

    @abc.abstractmethod
    def set_fps(self, fps: Number) -> bool:
        """
        Sets the fps for this camera

        :return: True on success, False otherwise
        """

    def rescale(self, factor: float) -> None:
        """
        Rescale the size of the frames read from this camera by a factor

        :param factor: The rescaling factor
        """
        self._set_width(int(self.get_width() * factor))
        self._set_height(int(self.get_height() * factor))
        self.get_data().focal_length *= factor

    def resize(self, fx: float, fy: float) -> None:
        """
        Rescale the size of the frames read from this camera by different width and height factors

        :param fx: The width factor
        :param fy: The height factor
        """
        self._set_width(int(self.get_width() * fx))
        self._set_height(int(self.get_height() * fy))
        self.get_data().focal_length *= np.sqrt(fx * fy)

    def set_frame_size(self, width: int, height: int) -> None:
        """
        Reset the width and height of frames read from this camera to given values
        
        :param width: The new width
        :param height: The new height
        """
        old_width, old_height = self.get_width(), self.get_height()
        self._set_height(height)
        self._set_width(width)
        self.get_data().focal_length *= np.sqrt(width * height / (old_width * old_height))



import struct
from math import ceil

from .udp_stream_broadcaster import UDPStreamBroadcaster
from gbvision.constants.net import UDP_MAX_SIZE, FRAGMENTED_UDP_HEADERS_STRUCT
from gbvision.constants.types import Number, Frame


class FragmentedUDPStreamBroadcaster(UDPStreamBroadcaster):
    """
    A UDP broadcaster that can fragment big frames and send them in chunks, thus allowing it to bypass the
    size limit of the UDP broadcaster

    :param mtu: The maximum size of a single fragment (without the fragment headers)
    """

    HEADER_SIZE = struct.calcsize(FRAGMENTED_UDP_HEADERS_STRUCT)

    def __init__(self, *args, mtu: Number = UDP_MAX_SIZE - HEADER_SIZE, **kwargs):
        UDPStreamBroadcaster.__init__(self, *args, **kwargs)
        assert self.HEADER_SIZE < mtu <= (UDP_MAX_SIZE - self.HEADER_SIZE), f"Invalid MTU value: {mtu}"
        self.mtu = mtu
        self.current_index = 0

    @staticmethod
    def _is_frame_legal_size(frame: bytes) -> bool:
        return True

    def _send_single_fragment(self, fragment: bytes, index: int, amount: int) -> None:
        """
        Sends the given fragment with the correct headers

        :param fragment: The fragment
        :param index: The fragment's index in the entire frame
        :param amount: The total amount of fragments that will be sent
        """
        assert len(fragment) <= self.mtu, f"Cannot send fragment of size {len(fragment)} when MTU is {self.mtu}"
        UDPStreamBroadcaster._send_bytes(self,
                                         struct.pack(FRAGMENTED_UDP_HEADERS_STRUCT, self.current_index, index,
                                                     amount) + fragment)

    def _send_bytes(self, data: Frame) -> None:
        fragments_needed = int(ceil(len(data) / self.mtu))
        for i in range(fragments_needed):
            self._send_single_fragment(data[i * self.mtu:min((i + 1) * self.mtu, len(data))], i, fragments_needed)
        self.current_index = (self.current_index + 1) & 0xffffffff



from .continues_shape import ContinuesShape
from gbvision.utils.shapes.base_circle import BaseCircle


class ContinuesCircle(ContinuesShape):
    """
    An implementation of ContinuesShape to circles.
    used to try and check whether two circles are indeed the same one.
    """

    @staticmethod
    def _base_shape():
        return BaseCircle




from typing import Tuple, Union

from .camera import Camera
from .camera_data import CameraData
from gbvision.constants.types import Frame
from gbvision.utils.net.stream_receiver import StreamReceiver


class StreamCamera(Camera):
    """
    A camera class which receives it's frames from a stream receiver

    :param data: The camera's CameraData object, should match the remote camera's data
    :param stream_receiver: The stream receiver to use
    """

    def __init__(self, data: CameraData, stream_receiver: StreamReceiver):
        self.__data = data.copy()
        self.stream_receiver = stream_receiver

    def release(self) -> None:
        self.stream_receiver.release()

    def is_opened(self) -> bool:
        return self.stream_receiver.is_opened()

    def set_exposure(self, exposure: Union[int, float, bool]) -> bool:
        return False

    def set_auto_exposure(self, auto: Union[int, float, bool]) -> bool:
        return False

    def get_data(self) -> CameraData:
        return self.__data

    def get_width(self) -> int:
        return self.stream_receiver.get_width()

    def get_height(self) -> int:
        return self.stream_receiver.get_height()

    def _set_width(self, width: int) -> None:
        self.stream_receiver.set_width(width)

    def _set_height(self, height: int) -> None:
        self.stream_receiver.set_height(height)

    def read(self) -> Tuple[bool, Frame]:
        return self.stream_receiver.read()

    def get_fps(self) -> int:
        return -1

    def set_fps(self, fps: int) -> bool:
        return False



from typing import List

import cv2

from gbvision.constants.math import SQRT_PI

from .base_shape import BaseShape
from gbvision.constants.types import Circle, Number, Point, Rect, Frame, Color


class BaseCircle(BaseShape):
    @staticmethod
    def set_center(shape: Circle, new_center: Point) -> Circle:
        return new_center, shape[1]

    @staticmethod
    def _unsafe_draw(frame: Frame, shape: Circle, color: Color, *args, **kwargs) -> None:
        cv2.circle(frame, (int(shape[0][0]), int(shape[0][1])), int(shape[1]), color, *args, **kwargs)

    @staticmethod
    def to_bounding_rect(shape: Circle) -> Rect:
        rect = (shape[0][0] - shape[1], shape[0][1] - shape[1], 2 * shape[1], 2 * shape[1])
        return rect

    @staticmethod
    def from_bounding_rect(bounding_rect: Rect) -> Circle:
        circle = (
            (bounding_rect[0] + bounding_rect[2] / 2, bounding_rect[1] + bounding_rect[3] / 2),
            (bounding_rect[3] + bounding_rect[2]) / 4)
        return circle

    from_contour = cv2.minEnclosingCircle

    @staticmethod
    def center(shape: Circle) -> Point:
        return shape[0]

    @staticmethod
    def collision(shape1: Circle, shape2: Circle) -> bool:
        center1, r1 = shape1
        center2, r2 = shape2
        return (center1[0] - center2[0]) ** 2 + (center1[1] - center2[1]) ** 2 < (r1 + r2) ** 2

    @classmethod
    def area(cls, shape: Circle) -> Number:
        return cls.root_area(shape) ** 2

    @classmethod
    def root_area(cls, shape: Circle) -> Number:
        return shape[1] * SQRT_PI

    @classmethod
    def sort(cls, shapes: List[Circle]):
        return sorted(shapes, key=cls.root_area)



from abc import ABC

from .window import Window
import cv2
from gbvision.models.system import EMPTY_PIPELINE
from gbvision.constants.types import ROI, Frame


class OpenCVWindow(Window, ABC):
    """
    A basic window that uses the opencv gui module

    :param exit_button: A list of chars (or string) that will close the window when pressed
    :param flags: OpenCV window flags (default is WINDOW_FREERATIO)
    """

    def __init__(self, window_name: str, exit_button='qQ', drawing_pipeline=EMPTY_PIPELINE, flags=cv2.WINDOW_FREERATIO):
        Window.__init__(self, window_name=window_name, drawing_pipeline=drawing_pipeline)
        self.exit_button = exit_button
        self.flags = flags
        self.last_key_pressed = None

    def _show_frame(self, frame: Frame) -> bool:
        if frame is None:
            return False
        cv2.imshow(self.window_name, frame)
        k = chr(cv2.waitKey(1) & 0xFF)
        self.last_key_pressed = k
        if k in self.exit_button:
            return False
        return True

    def _open(self) -> None:
        cv2.namedWindow(self.window_name, self.flags)

    def _release(self) -> None:
        cv2.destroyWindow(self.window_name)

    def select_roi(self, frame: Frame) -> ROI:
        """
        Presents this frame on the window, and allows the user to select a rectangular area from the frame

        :param frame: The frame to show
        :return: The selected rectangular area, as a gbvision.ROI (similar to gbvision.Rect but all values are integers)
        """
        return cv2.selectROI(self.window_name, frame)



from typing import Tuple

from gbvision.constants.types import Frame
from .streaming_camera import SimpleStreamingCamera
from gbvision.models.cameras import UNKNOWN_CAMERA
from .usb_camera import USBCamera


class USBStreamingCamera(SimpleStreamingCamera, USBCamera):
    """
    A simple USB stream camera
    """

    def release(self) -> None:
        USBCamera.release(self)
        self.stream_broadcaster.release()

    def _read(self) -> Tuple[bool, Frame]:
        return USBCamera.read(self)

    def __init__(self, broadcaster, port, should_stream=False, data=UNKNOWN_CAMERA):
        SimpleStreamingCamera.__init__(self, broadcaster, should_stream=should_stream)
        USBCamera.__init__(self, port, data)



import numpy as np
import cv2

from .base_shape import BaseShape
from gbvision.constants.types import Line, Number, Point, Frame, Color, Rect, Contour
from gbvision.tools.math_tools import almost_equal


class BaseLine(BaseShape):
    @classmethod
    def set_center(cls, shape: Line, new_center: Point) -> Line:
        current_center = cls.center(shape)
        center_diff = new_center[0] - current_center[0], new_center[1] - current_center[1]
        return (shape[0][0] + center_diff[0], shape[0][1] + center_diff[1]), (
            shape[1][0] + center_diff[0], shape[1][1] + center_diff[1])

    @staticmethod
    def _unsafe_draw(frame: Frame, shape: Line, color: Color, *args, **kwargs) -> None:
        cv2.line(frame, (int(shape[0][0]), int(shape[0][1])), (int(shape[1][0]), int(shape[1][1])), color, *args,
                 **kwargs)

    @staticmethod
    def to_bounding_rect(shape: Line) -> Rect:
        return min(shape[0][0], shape[1][0]), min(shape[0][1], shape[1][1]), abs(shape[0][0] - shape[1][0]), abs(
            shape[0][1] - shape[1][1])

    @staticmethod
    def from_bounding_rect(bounding_rect: Rect) -> Line:
        # There is no way of knowing if the line was top-left to bottom-right or top-right to bottom-left
        return NotImplemented

    @staticmethod
    def from_contour(cnt: Contour) -> Line:
        # Cant convert a contour to a line
        return NotImplemented

    @staticmethod
    def center(shape: Line) -> Point:
        return (shape[0][0] + shape[1][0]) / 2, (shape[0][1] + shape[1][1]) / 2

    @staticmethod
    def angle(shape: Line) -> Number:
        """
        Calculates the angle at (x0 + 1, y0), (x0, y0), (x1, y1)

        :param shape: The line who's angle is needed
        :return: The angle of the line (between -pi and pi)
        """
        return np.arctan2(shape[0][1] - shape[1][1], shape[0][0] - shape[1][0])

    @classmethod
    def angle_positive(cls, shape: Line) -> Number:
        """
        Calculates the angle at (x0 + 1, y0), (x0, y0), (x1, y1)
        Will always return a positive value

        :param shape: The line who's angle is needed
        :return: The angle of the line (between 0 and 2*pi)
        """
        angle = cls.angle(shape)
        if angle < 0:
            angle = 2 * np.pi + angle
        return angle

    @staticmethod
    def infinite_line_params(shape: Line) -> Point:
        """
        Calculates the parameters (m, b) such that the function y = m * x + b contains the entire given line

        :param shape: The line to find
        :return: A tuple of (m, b)
        """
        m = (shape[0][1] - shape[1][1]) / (shape[0][0] - shape[1][0])
        b = shape[0][1] - m * shape[0][0]
        return m, b

    @classmethod
    def collision(cls, shape1: Line, shape2: Line) -> bool:
        if almost_equal(shape1[0][0], shape1[1][0]):  # parallel to y
            if almost_equal(shape2[0][1], shape2[1][1]):  # parallel to x
                return min(shape2[0][0], shape2[1][0]) <= shape1[0][0] <= max(shape2[0][0], shape2[1][0]) and \
                       min(shape1[0][1], shape1[1][1]) <= shape2[0][1] <= max(shape1[0][1], shape1[1][1])
            else:  # inverse x and y
                shape1 = (shape1[0][1], -shape1[0][0]), (shape1[1][1], -shape1[1][0])
                shape2 = (shape2[0][1], -shape2[0][0]), (shape2[1][1], -shape2[1][0])
        if almost_equal(shape2[0][0], shape2[1][0]):  # parallel to y
            if almost_equal(shape1[0][1], shape1[1][1]):  # parallel to x
                return min(shape1[0][0], shape1[1][0]) <= shape2[0][0] <= max(shape1[0][0], shape1[1][0]) and \
                       min(shape2[0][1], shape2[1][1]) <= shape1[0][1] <= max(shape2[0][1], shape2[1][1])
            else:  # rotate by 90 deg
                shape1 = (shape1[0][1], -shape1[0][0]), (shape1[1][1], -shape1[1][0])
                shape2 = (shape2[0][1], -shape2[0][0]), (shape2[1][1], -shape2[1][0])

        m1, b1 = cls.infinite_line_params(shape1)
        m2, b2 = cls.infinite_line_params(shape2)
        if almost_equal(m1, m2):
            if not almost_equal(b1, b2):
                return False
            return min(shape1[0][0], shape1[1][0]) <= max(shape2[0][0], shape2[1][0]) <= max(shape1[0][0],
                                                                                             shape1[1][0]) or min(
                shape2[0][0], shape2[1][0]) <= max(shape1[0][0], shape1[1][0]) <= max(shape2[0][0], shape2[1][0])

        collision_x = (b1 - b2) / (m2 - m1)
        return min(shape1[0][0], shape1[1][0]) <= collision_x <= max(shape1[0][0], shape1[1][0]) and \
               min(shape2[0][0], shape2[1][0]) <= collision_x <= max(shape2[0][0], shape2[1][0])

    @classmethod
    def area(cls, shape: Line) -> Number:
        # A line has no area
        return 0



import socket
from .stream_broadcaster import StreamBroadcaster
from gbvision.constants.net import UDP_MAX_SIZE


class UDPStreamBroadcaster(StreamBroadcaster):
    """
    this class uses UDP to send a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the client and the receiver is the server
    WARNING: Do not use this class to send large images, UDP has a limited packet size (the IPv4 limit)
    To send a stream of frames larger then the maximum size of an IP packet, use FragmentedUDPStreamBroadcaster

    :param ip: The IPv4 address of the udp stream receiver, for example '10.45.90.5'
    :param port: The UDP port to use
    """

    def __init__(self, ip: str, port: int, *args, **kwargs):
        StreamBroadcaster.__init__(self, *args, **kwargs)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.server_addr = (ip, port)

    @staticmethod
    def _is_frame_legal_size(data: bytes) -> bool:
        """
        Checks if the data's size is too large for a single UDP packet

        :param data: The data to try and send
        :return: True if the data is small enough, False otherwise
        """
        return len(data) < UDP_MAX_SIZE

    def _can_send_bytes(self, data: bytes) -> bool:
        return self._is_frame_legal_size(data) and StreamBroadcaster._can_send_bytes(self, data)

    def _send_bytes(self, data: bytes) -> None:
        self.socket.sendto(data, self.server_addr)

    def release(self) -> None:
        self.socket.close()

    def is_opened(self) -> bool:
        return self.socket.fileno() != -1



from gbvision.constants.types import Number


def almost_equal(x: Number, y: Number, delta: Number = 0.001) -> bool:
    """
    Checks if two numbers are almost equal using the formula -delta <= x - y <= delta

    :param x: The first number
    :param y: The second number
    :param delta: The maximum allowed diff (default 0.001)
    :return: True if the numbers are almost equal, False otherwise
    """
    return -delta <= x - y <= delta



import abc

from gbvision.utils.releasable import Releasable
from gbvision.constants.types import Frame


class Recorder(Releasable, abc.ABC):
    """
    An abstract recorder class
    Records a given feed of frames into a file

    :param file_name: the file name
    """
    def __init__(self, file_name):
        self.file_name = file_name

    @abc.abstractmethod
    def write(self, frame: Frame) -> Frame:
        """
        Writes the given frame to the file

        :param frame: the frame to write
        """



import numpy as np
import cv2

from .base_polygon import BasePolygon
from gbvision.constants.types import Polygon, Contour


class BaseConvexPolygon(BasePolygon):
    @staticmethod
    def collision(shape1: Polygon, shape2: Polygon) -> bool:
        shape1, shape2 = np.array(shape1), np.array(shape2)
        shape1 = shape1.reshape(shape1.size // 2, 2)
        shape2 = shape2.reshape(shape1.size // 2, 2)

        projection_axises = []

        negation_array = np.array([-1.0, 1.0])

        for i in range(len(shape1)):
            projection_axises.append(shape1[(i + 1) % len(shape1)] - shape1[i])
        for i in range(len(shape2)):
            projection_axises.append(shape2[(i + 1) % len(shape2)] - shape2[i])

        for i in range(len(projection_axises)):
            projection_axises[i] = projection_axises[i][::-1] * negation_array

        for axis in projection_axises:
            proj1 = []
            for vertix in shape1:
                proj1.append(vertix.dot(axis))
            proj1 = min(proj1), max(proj1)

            proj2 = []
            for vertix in shape2:
                proj2.append(vertix.dot(axis))
            proj2 = min(proj2), max(proj2)

            if proj1[0] > proj2[1] or proj2[0] > proj1[1]:
                return False

        return True

    @staticmethod
    def from_contour(cnt: Contour, arc_length_multiplier=0.05) -> Polygon:
        polygon = BasePolygon.from_contour(cnt, arc_length_multiplier)
        return cv2.convexHull(polygon)



from copy import deepcopy

import numpy as np

from gbvision.constants.types import Number, Point, Frame, Color, Rect, Contour
from .base_circle import BaseCircle
from .base_rect import BaseRect
from .base_shape import BaseShape


class BasePoint(BaseShape):
    @staticmethod
    def set_center(shape: Point, new_center: Point) -> Point:
        return deepcopy(new_center)

    @staticmethod
    def _unsafe_draw(frame: Frame, shape: Point, color: Color, *args, **kwargs) -> None:
        BaseCircle.draw(frame, (shape, 1), color, *args, **kwargs)

    @staticmethod
    def to_bounding_rect(shape: Point) -> Rect:
        return shape[0] - 1, shape[1] - 1, 2, 2

    @staticmethod
    def from_bounding_rect(bounding_rect: Rect) -> Point:
        return BaseRect.center(bounding_rect)

    @staticmethod
    def from_contour(cnt: Contour) -> Point:
        # Cannot convert a contour to a point
        return NotImplemented

    @staticmethod
    def center(shape: Point) -> Point:
        return deepcopy(shape)

    @staticmethod
    def collision(shape1: Point, shape2: Point) -> bool:
        return shape1[0] == shape2[0] and shape1[1] == shape2[1]

    @classmethod
    def area(cls, shape: Point) -> Number:
        # A point has no area
        return 0

    @staticmethod
    def rotate(shape: Point, angle: Number) -> Point:
        """
        Rotates the point around the (0, 0) point by the given angle

        :param shape: The point to rotate
        :param angle: The angle by which to rotate
        :return: The point after rotation
        """
        sin = np.sin(angle)
        cos = np.cos(angle)
        return cos * shape[0] - sin * shape[1], cos * shape[1] + sin * shape[0]



import cv2
import numpy as np

from gbvision.constants.math import EPSILON
import gbvision.utils.cameras as cameras
from gbvision.constants.types import Point, Contour, Number, Location


class GameObject:
    """
    constructor of the image object
    which is an object on field

    :param root_area: the square root of the surface area of the object in real life

    """

    def __init__(self, root_area: Number):
        self.root_area = root_area

    def distance(self, camera: cameras.Camera, root_area: Number) -> float:
        """
        Note: this measures the distance between the camera and the object, to use another measuring point
        calculate the norm of the location

        :param camera: the camera, can be either Camera or CameraList
        :param root_area: a float representing the square root of the area of the object
            (in pixels)
        :return: the norm of the vector between the camera and the object (in meters)
        """
        return camera.get_data().focal_length * self.root_area / (root_area + EPSILON)

    def location(self, camera: cameras.Camera, root_area: Number, center: Point) -> Location:
        """
        Calculates the 3D location of this object, relative to the camera

        :param camera: the camera, can be either Camera or CameraList
        :param root_area: a float representing the square root of the area of the object (in pixels)
        :param center: the center (x,y) of this object in the frame
        :return: a 3d vector of the relative [x y z] location between the object and the camera/measuring point (in meters)
        """
        frame_center = camera.get_width(), camera.get_height()
        frame_center = np.array(frame_center) / 2
        x, y = np.array(center) - frame_center
        alpha = x * camera.get_data().fov_width / frame_center[0]
        beta = y * camera.get_data().fov_height / frame_center[1]
        rel = np.array([[np.sin(alpha), np.sin(beta),
                         np.sqrt(1 - np.sin(alpha) ** 2 - np.sin(beta) ** 2)]]) * self.distance(camera, root_area)
        return camera.get_data().rotation_matrix.dot(rel.T).flatten() + camera.get_data().offset



import abc
from typing import Any, Tuple

from gbvision.constants.types import Frame
from .window import Window
from gbvision.models.system import EMPTY_PIPELINE
from threading import Thread

from gbvision.utils.readable import Readable


class WrapperWindow(Readable, Window, abc.ABC):
    """
    A window class that uses an object (wrap_object) and reads frames from it in to display it's feed
    The window is also a readable, calling "read" will show a frame and return it

    :param wrap_object: An object to read frames from, can be of any type
    """

    def __init__(self, window_name: str, wrap_object: Any, drawing_pipeline=EMPTY_PIPELINE):
        Window.__init__(self, window_name=window_name, drawing_pipeline=drawing_pipeline)
        self.wrap_object = wrap_object
        self.__width = 0
        self.__height = 0

    def read(self) -> Tuple[bool, Frame]:
        """
        Shows one frame and returns it

        :return: The frame if the window was not closed, None otherwise
        """
        ok, frame = self._get_frame()
        if ok and self.show_frame(frame):
            self.__width = frame.shape[1]
            self.__height = frame.shape[0]
            return True, frame
        return False, None

    def get_width(self) -> int:
        return self.__width

    def get_height(self) -> int:
        return self.__height

    def show(self) -> None:
        """
        Reads from the wrap object and shows the frame until the window is closed
        """
        while True:
            if not self.read()[0]:
                return

    @abc.abstractmethod
    def _get_frame(self) -> Tuple[bool, Frame]:
        """
        Unsafely reads a frame from the wrapped object and returns the read frame
        
        :return: A tuple of two values, the first is a boolean indicating if the method has succeded, the second is
                 the read frame (or None on failure)
        """

    def show_async(self) -> None:
        """
        Opens the steam video window on another thread
        """
        # Close existing window to allow it to re-open on the new thread
        if self.is_opened():
            self.release()
        Thread(target=self.show).start()



import abc
import pickle
from typing import Optional

import cv2

from gbvision.constants.types import Frame, Coordinates
from gbvision.utils.readable import Readable


class StreamReceiver(Readable, abc.ABC):
    """
    this is an abstract receiver that receives stream from a broadcast receiver
    this class should not be instanced but inherited from

    :param shape: Optional. The shape (x, y) of the sent frame (None to use the given frame's shape)
    :param fx: Optional, The ratio between width of the read frame to the width of the frame returned (None for 1)
    :param fy: ratio between height of the read frame to the height of the frame returned (None for 1)
    :param convert_from_grayscale: Should this receiver convert grayscale images to BGR images (default True)
    """

    def __init__(self, shape: Optional[Coordinates] = None, fx: Optional[float] = None, fy: Optional[float] = None,
                 convert_from_grayscale: bool = True):
        self.shape = shape
        self.fx = fx
        self.fy = fy
        self.convert_from_grayscale = convert_from_grayscale
        self.__width = 0
        self.__height = 0

    @abc.abstractmethod
    def _get_bytes(self) -> bytes:
        """
        performs the entire process of reading from the socket
        any defragmentation and headers handling necessary needs to happen here

        :return: the bytes read from the socket, after defragmentation (if exists)
        """

    def read(self):
        frame_data = self._get_bytes()
        frame = self._from_bytes(frame_data)
        if frame is None:
            return False, None
        frame = cv2.imdecode(frame, -1)
        return True, self._prep_frame(frame)

    @staticmethod
    def _from_bytes(bytes_obj: bytes) -> object:
        """
        parses a binary represented object back to a python object

        :param bytes_obj: the binary output from the socket
        :return: a python object
        """
        return pickle.loads(bytes_obj)

    def _prep_frame(self, frame: Frame) -> Frame:
        """
        prepares the frame to be returned and used
        resize and convert to bgr channeled image
        
        :param frame: the frame to be prepared
        :return: the frame after preparations
        """
        if frame is None:
            return frame
        if len(frame.shape) < 3 and self.convert_from_grayscale:
            frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)
        if self.shape is not None:
            frame = cv2.resize(frame, self.shape)
        if self.fx is not None or self.fy is not None:
            frame = cv2.resize(frame, (0, 0), fx=self.fx or 1.0, fy=self.fy or 1.0)
        self.__height, self.__width = frame.shape[:2]
        return frame

    def get_width(self) -> int:
        if self.shape is not None:
            return self.shape[0]
        return self.__width

    def get_height(self) -> int:
        if self.shape is not None:
            return self.shape[1]
        return self.__height

    def set_width(self, width: int) -> None:
        self.shape = (width, self.get_height())

    def set_height(self, height: int) -> None:
        self.shape = (self.get_width(), height)



from .basic_ops import gray, sharpen, blue, blur, corners, red, edges, green, normalize, distance_transform, \
    normalized_distance_transform
from .cameras import LIFECAM_3000, UNKNOWN_CAMERA
from .contours import fix_contours_shape, convex_hull, convex_hull_multiple, find_contours, FilterContours
from .system import EMPTY_PIPELINE, EMPTY_GAME_OBJECT



import typing
from typing import Callable, Tuple, Optional

import cv2

from gbvision.constants.types import Frame, Rect, TrackerType

_major_ver, _minor_ver, _subminor_ver = cv2.__version__.split('.')


class _EmptyTracker:
    def __init__(self):
        self.__rect = None

    def init(self, frame: Frame, rect: Rect):
        self.__rect = rect
        return True

    def update(self, frame: Frame) -> Tuple[bool, Rect]:
        return True, self.__rect


def _get_tracker_create(tracker_name: str) -> Optional[Callable[[], TrackerType]]:
    if tracker_name == 'EMPTY':
        return _EmptyTracker
    try:
        if int(_major_ver) < 4 and int(_minor_ver) < 3:
            return lambda: cv2.cv2.Tracker_create(tracker_name.upper())
        attr_full_name = f'Tracker{tracker_name}_create'
        if hasattr(cv2, attr_full_name):
            return cv2.__dict__[attr_full_name]
        return cv2.legacy.__dict__[attr_full_name]
    except AttributeError:
        return None


class Tracker:
    """
    A tracker object that tracks a rectangle in a video using an opencv tracking algorithm

    :param tracker_type: Tracker algorithm taken from this list: BOOSTING, MIL, KCF, TLD, MEDIANFLOW,
        GOTURN, MOSSE, CSRT, EMPTY. (Default is EMPTY)
    """

    TRACKER_TYPE_BOOSTING = 'BOOSTING'
    TRACKER_TYPE_MIL = 'MIL'
    TRACKER_TYPE_KCF = 'KCF'
    TRACKER_TYPE_TLD = 'TLD'
    TRACKER_TYPE_MEDIANFLOW = 'MEDIANFLOW'
    TRACKER_TYPE_GOTURN = 'GOTURN'
    TRACKER_TYPE_MOSSE = 'MOSSE'
    TRACKER_TYPE_CSRT = 'CSRT'
    TRACKER_TYPE_EMPTY = 'EMPTY'

    _TRACKER_ALGORITHMS = {
        tracker_name: _get_tracker_create(tracker_type)
        for tracker_name, tracker_type in
        [(TRACKER_TYPE_BOOSTING, 'Boosting'),
         (TRACKER_TYPE_MIL, 'MIL'),
         (TRACKER_TYPE_KCF, 'KCF'),
         (TRACKER_TYPE_TLD, 'TLD'),
         (TRACKER_TYPE_MEDIANFLOW, 'MedianFlow'),
         (TRACKER_TYPE_GOTURN, 'GOTURN'),
         (TRACKER_TYPE_MOSSE, 'MOSSE'),
         (TRACKER_TYPE_CSRT, 'CSRT'),
         (TRACKER_TYPE_EMPTY, 'EMPTY')]
    }

    def __init__(self, tracker_type: str = TRACKER_TYPE_EMPTY):
        tracker_type = tracker_type.upper()
        assert tracker_type in self._TRACKER_ALGORITHMS, f'Unknown tracker type: {tracker_type}'
        assert self._TRACKER_ALGORITHMS[tracker_type] is not None,\
            f'Your version of OpenCV has no support for tracker type {tracker_type}'
        self.tracker = self._TRACKER_ALGORITHMS[tracker_type]()
        self.tracker_type = tracker_type

    def init(self, frame: Frame, rect: Rect) -> bool:
        """
        Initialize the tracker

        :param frame: The frame
        :param rect: Given rectangle
        :return: True if initialization went successfully, False otherwise
        """
        return self.tracker.init(frame, typing.cast(Rect, tuple([int(max(x, 0)) for x in rect])))

    def update(self, frame: Frame) -> Rect:
        """
        Get the rect location in new frame

        :param frame: The frame
        :return: The location of the rect in new frame
        """
        return self.tracker.update(frame)[1]



import struct
import time
from typing import Optional, Tuple

import typing

from .udp_stream_receiver import UDPStreamReceiver
from gbvision.constants.net import FRAGMENTED_UDP_HEADERS_STRUCT
from gbvision.constants.types import Number


class FragmentedUDPStreamReceiver(UDPStreamReceiver):
    """
    A UDP receiver that can de-fragment packets sent from the fragmented UDP broadcaster, and rebuild large
    frames from several chunks

    :param max_wait: The maximum time in seconds the receiver will wait for chunks of a frame to arrive from
    the moment it received the first chunk, if this time passed and not all fragments have arrived,
    the receiver will stop handling the current frame and move on the the next frame that will arrive
    can be None to not set a wait limit (same as infinity)
    If this value is too large, a single frame might block the entire receiver if not all fragments arrive
    If this value is too small, a frame might be dropped just because this time passed before all fragments
    could arrive
    """

    def __init__(self, *args, max_wait: Optional[Number] = 1, **kwargs):
        UDPStreamReceiver.__init__(self, *args, **kwargs)
        self.headers_size = struct.calcsize(FRAGMENTED_UDP_HEADERS_STRUCT)
        assert (max_wait is None) or (max_wait > 0), f"Invalid max wait value: {max_wait}"
        self.max_wait = max_wait

    def _read_single_fragment(self) -> Tuple[Tuple[int, int, int], bytes]:
        """
        Reads a single fragment from the UDP socket

        :return: A tuple of two values:
                 1. The fragment headers as a tuple of three values (frame_index, fragment_index, fragments_amount)
                 2. The fragment payload
        """
        fragment = UDPStreamReceiver._get_bytes(self)
        headers = fragment[:self.headers_size]
        fragment = fragment[self.headers_size:]
        return typing.cast(Tuple[int, int, int], struct.unpack(FRAGMENTED_UDP_HEADERS_STRUCT, headers)), fragment

    def _get_bytes(self) -> bytes:
        current_id = -1
        fragments = {}
        last_time = 0
        while True:
            (new_id, fragment_index, amount), fragment = self._read_single_fragment()
            if new_id > current_id:
                last_time = time.time()
                current_id = new_id
                fragments.clear()
            if new_id < current_id:
                continue
            fragments[fragment_index] = fragment
            if len(fragments) == amount:
                break
            if self.max_wait is not None and time.time() - last_time > self.max_wait:
                current_id = -1
        # we want the fragments in their correct order, not necessarily the order they arrived
        return b''.join([fragments[i] for i in range(len(fragments))])



import abc

from gbvision.utils.releasable import Releasable
from gbvision.models.system import EMPTY_PIPELINE
from gbvision.constants.types import Frame


class Window(Releasable, abc.ABC):
    """
    An abstract window class

    :param window_name: The title of the new window
    :param drawing_pipeline: Optional. A pipeline that draws on each frame before displaying it
    """

    def __init__(self, window_name: str, drawing_pipeline=EMPTY_PIPELINE):
        """
        initializes the window
        """
        self.window_name = window_name
        self._is_opened = False
        self.drawing_pipeline = drawing_pipeline

    @abc.abstractmethod
    def _show_frame(self, frame: Frame) -> bool:
        """
        Shows the frame
        :param frame: the frame to show
        :return: False if the window should be closed, True otherwise
        """

    def is_opened(self) -> bool:
        """
        Checks if the window is opened

        :return: True is the window is opened, False otherwise
        """
        return self._is_opened

    def show_frame(self, frame: Frame) -> bool:
        """
        Shows the frame on the window

        :param frame: the frame to show
        :return: false if the window was closed, true otherwise
        """
        if not self.is_opened():
            self.open()
        if self._show_frame(self.drawing_pipeline(frame)):
            return True
        self.release()
        return False

    def show_and_return_frame(self, frame: Frame) -> Frame:
        """
        Shows and returns the given frame

        :param frame: The frame to show
        :return: The given frame, or None if the window has closed
        """
        return frame if self.show_frame(frame) else None

    @abc.abstractmethod
    def _open(self) -> None:
        """
        Unsafely opens the window
        Not to be used by the programmer, only by the function open
        """

    @abc.abstractmethod
    def _release(self) -> None:
        """
        Unsafely closes the window
        Not to be used by the programmer, only by the function close
        """

    def open(self) -> None:
        """
        Opens the window
        """
        if not self.is_opened():
            self._open()
            self._is_opened = True

    def release(self) -> None:
        """
        Closes the window
        """
        if self.is_opened():
            self._release()
            self._is_opened = False



import abc
from typing import List, Iterable, Union, Callable

from gbvision.constants.types import Frame, Location, Shape, FilterFunction, Contour, Number
from gbvision.models.contours import find_contours
from gbvision.models.system import EMPTY_PIPELINE, EMPTY_GAME_OBJECT
from gbvision.utils.cameras.camera import Camera
from gbvision.utils.game_object import GameObject
from gbvision.utils.shapes.base_shape import BaseShapeType
from gbvision.utils.thresholds.threshold import Threshold


class ShapeFinder(abc.ABC):
    """
    This is an abstract class that represents an object finder
    An object finder is a type that outputs an object's 3d real location based on an of it image it's
    GameObject real-life parameters

    :param threshold_func: The threshold function used to find the shapes in the frame
    :param game_object: The game object descriptor for the real-life parameters of the finder's target
    :param contours_hook: Optional. A pipeline to run on the contours after finding them
    :param shapes_hook: Optional. A pipeline to run on the shapes after parsing the contours to them
    """

    def __init__(
            self,
            threshold_func: Union[FilterFunction, Threshold],
            game_object: GameObject = EMPTY_GAME_OBJECT,
            contours_hook: Callable[[List[Contour]], List[Contour]] = EMPTY_PIPELINE,
            shapes_hook: Callable[[List[Shape]], List[Shape]] = EMPTY_PIPELINE,
    ):
        self.game_object = game_object
        self._find_shapes_pipeline = EMPTY_PIPELINE + \
                                     threshold_func + \
                                     find_contours + \
                                     contours_hook + \
                                     self._base_shape().from_contours + \
                                     shapes_hook

    def __call__(self, frame: Frame, camera: Camera) -> List[Location]:
        """
        Finds all instances of the object in the frame

        :param frame: The frame in which to find
        :param camera: The camera used to read the frame
        :return: All object of this type in the physical space
        """
        return self.locations_from_shapes(self.find_shapes(frame), camera)

    @staticmethod
    @abc.abstractmethod
    def _base_shape() -> BaseShapeType:
        """
        Returns the base shape matching this finder

        :return: The base shape (a class that inherits from BaseShape)
        """

    def find_shapes_unsorted(self, frame: Frame) -> List[Shape]:
        """
        Finds all the shapes and returns them in frame after full pipeline (not sorted)

        :param: The current frame the finder searches in
        :return: A list of shapes: see gbvision/constants/types
        """
        return self._find_shapes_pipeline(frame)

    def find_shapes(self, frame: Frame) -> List[Shape]:
        """
        Finds all the shapes and returns them in frame after full pipeline (sorted)

        :param: The current frame the finder searches in
        :return: A list of shapes: see gbvision/constants/types
        """
        return self._base_shape().sort(self.find_shapes_unsorted(frame))

    def find_and_filter_shapes(self, frame: Frame) -> List[Shape]:
        """
        Finds all the shapes and returns them in frame after full pipeline (sorted), removes all inner shapes

        :param: The current frame the finder searches in
        :return: A list of shapes: see gbvision/constants/types
        """
        return self._base_shape().filter_inners(self.find_shapes(frame))

    def locations_from_shapes(self, shapes: Iterable[Shape], camera: Camera) -> List[Location]:
        """
        Finds the locations of the shapes based on the shape descriptor and camera constants

        :param shapes: A list of the shapes
        :param camera: The camera used to capture the frame that the shapes were found in
        :return: A list of the locations of all the shapes
        """
        return list(
            map(lambda shape: self.game_object.location(camera,
                                                        self._base_shape().root_area(shape),
                                                        self._base_shape().center(shape)), shapes))



from typing import List

import cv2

from gbvision.constants.system import CONTOURS_INDEX
from gbvision.constants.types import Contour, Polygon, Frame
from gbvision.utils.pipeline import PipeLine


@PipeLine
def find_contours(frame: Frame) -> List[Contour]:
    """
    Finds the contours in a binary frame

    :param frame: The frame (usually after threshold and denoising)
    :return: A list of all the contours in the frame
    """
    # DO NOT CHANGE THE CHAIN_APPROX_NONE
    # You do not know the damages it may cause
    return cv2.findContours(frame, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)[CONTOURS_INDEX]


class FilterContours(PipeLine):
    """
    A pipeline factory that receives a list of contours and filters out all contours with an area less then defined
    to the pipeline

    :param min_area: the minimal area of a contour in order for it to pass the filter
    """

    def __init__(self, min_area: float):
        PipeLine.__init__(self, lambda cnts: filter(lambda c: cv2.contourArea(c) >= min_area, cnts), list)


convex_hull = PipeLine(cv2.convexHull)

convex_hull_multiple = PipeLine(lambda x: list(map(convex_hull, x)))


# SHAPES

@PipeLine
def fix_contours_shape(cnts: List[Contour]) -> List[Polygon]:
    """
    fixes the contours to a usable shape
    the shape of the contours is a list of tuples of integers/floats, where each tuple is a point
    an example of two rectangles represented with this shape will be:
    [[(0, 0), (0, 2), (1, 2), (1, 0)],
    [(5, 4), (7, 4), (7, 9), (9, 5)]]
    
    :param cnts: the contours / polygons list whose shape should be fixed
    :return: a list of all the contours with the fixed shape
    """
    cnts = map(lambda polydp: map(lambda x: x[0], polydp), cnts)
    return list(map(lambda polydp: list(map(tuple, polydp)), cnts))



from .object_finder import ShapeFinder
from gbvision.utils.shapes.base_rect import BaseRect
from ..shapes.base_shape import BaseShapeType


class RectFinder(ShapeFinder):
    """
    Finds a rectangular shaped object
    """

    @staticmethod
    def _base_shape() -> BaseShapeType:
        return BaseRect



from .object_finder import ShapeFinder
from gbvision.utils.shapes.base_convex_polygon import BaseConvexPolygon


class ConvexPolygonFinder(ShapeFinder):
    """
    Finds any generic convex polygon, not recommended when another finder can be used
    """

    @staticmethod
    def _base_shape():
        return BaseConvexPolygon



import abc


class Releasable(abc.ABC):
    """
    An interface representing an object holding some resources, which can be released
    """

    @abc.abstractmethod
    def release(self) -> None:
        """
        Closes this releasable and releases it's resource
        """

    @abc.abstractmethod
    def is_opened(self) -> bool:
        """
        Checks if this releasable is open and can be read from

        :return: True if the releasable is opened, False otherwise
        """

    def __enter__(self) -> 'Releasable':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self.is_opened():
            self.release()

    def __del__(self) -> None:
        if self.is_opened():
            self.release()



from .continues_shape import ContinuesShape
from gbvision.utils.shapes.base_rotated_rect import BaseRotatedRect


class ContinuesRotatedRect(ContinuesShape):
    """
    An implementation of ContinuesShape to rotated rectangles.
    used to try and check whether two rotated rectangles are indeed the same one.
    """

    @staticmethod
    def _base_shape():
        return BaseRotatedRect
